"""
References
----------
    Allen, R. and L. Pereira (2009).  Estimating crop coefficients from
    fraction of ground cover and height. Irrigation Science 28:17-34.
    DOI 10.1007/s00271-009-0182-z
"""

# Scaling factor to use in EE remap function since it only work with integers
int_scalar = 100

cdl = {
    # Field crops
    49: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Onions'},
    206: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Carrots'},
    208: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Garlic'},
    214: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Broccoli'},
    227: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Lettuce'},
    243: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Cabbage'},
    244: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Cauliflower'},
    245: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Celery'},
    246: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Radishes'},

    # Vegetables, solanum family
    54: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Tomatoes'},
    248: {'crop_class': 1, 'h_max': 0.8, 'm_l': 2, 'fr_mid': 1, 'name': 'Eggplants'},

    # Vegetables, cucurb family
    48: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Watermelons'},
    209: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Cantaloupes'},
    213: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Honeydew Melons'},
    222: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Squash'},
    228: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Dbl Crop Triticale/Corn'},
    229: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Pumpkins'},

    # Vegetables, other
    47: {'crop_class': 1, 'h_max': 0.37, 'm_l': 2, 'fr_mid': 1, 'name': 'Misc Vegs & Fruits'},

    # Roots & tubers
    41: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Sugarbeets'},
    43: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Potatoes'},
    46: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Sweet Potatoes'},
    247: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Turnips'},

    # Legumes
    5: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Soybeans'},
    10: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Peanuts'},
    30: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Speltz'},
    42: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Dry Beans'},
    51: {'crop_class': 1, 'h_max': 0.4, 'm_l': 2, 'fr_mid': 1, 'name': 'Chick Peas'},
    52: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Lentils'},
    53: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Peas'},
    224: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Vetch'},

    # Perennial vegetables or fruit
    14: {'crop_class': 1, 'h_max': 0.7, 'm_l': 2, 'fr_mid': 1, 'name': 'Mint'},
    50: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Cucumbers'},
    207: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Asparagus'},
    216: {'crop_class': 1, 'h_max': 0.7, 'm_l': 2, 'fr_mid': 1, 'name': 'Peppers'},
    221: {'crop_class': 1, 'h_max': 0.2, 'm_l': 2, 'fr_mid': 1, 'name': 'Strawberries'},

    # Fiber
    2: {'crop_class': 1, 'h_max': 1.35, 'm_l': 2, 'fr_mid': 1, 'name': 'Cotton'},
    6: {'crop_class': 1, 'h_max': 2.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Sunflower'},
    31: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Canola'},
    32: {'crop_class': 1, 'h_max': 1.2, 'm_l': 2, 'fr_mid': 1, 'name': 'Flaxseed'},
    33: {'crop_class': 1, 'h_max': 0.8, 'm_l': 2, 'fr_mid': 1, 'name': 'Safflower'},
    45: {'crop_class': 1, 'h_max': 3.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Sugarcane'},

    # Cereal
    1: {'crop_class': 1, 'h_max': 2.0, 'm_l': 2, 'fr_mid': 1,  'name': 'Corn'},
    4: {'crop_class': 1, 'h_max': 1.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Sorghum'},
    12: {'crop_class': 1, 'h_max': 1.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Sweet Corn'},
    21: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Barley'},
    22: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Durum Wheat'},
    23: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Spring Wheat'},
    24: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Winter Wheat'},
    25: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Other Small Grains'},
    26: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Dbl Crop WinWht/Soybeans'},
    28: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Oats'},
    29: {'crop_class': 1, 'h_max': 1.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Millet'},
    # CGM - Repeat of crop 32 in fiber section
    #   Same values so commenting out for now
    # 32: {'crop_class': 1, 'h_max': 1.2, 'm_l': 2, 'fr_mid': 1, 'name': 'Flaxseed'},
    34: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Rape Seed'},
    35: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Mustard'},
    39: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Buckwheat'},
    56: {'crop_class': 1, 'h_max': 5.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Hops'},

    # Rice
    3: {'crop_class': 5, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Rice'},

    # Forage
    27: {'crop_class': 1, 'h_max': 0.3, 'm_l': 2, 'fr_mid': 1, 'name': 'Rye'},
    36: {'crop_class': 1, 'h_max': 0.7, 'm_l': 2, 'fr_mid': 1, 'name': 'Alfalfa'},
    58: {'crop_class': 1, 'h_max': 0.6, 'm_l': 2, 'fr_mid': 1, 'name': 'Clover/Wildflowers'},
    37: {'crop_class': 1, 'h_max': 0.5, 'm_l': 2, 'fr_mid': 1, 'name': 'Other Hay/Non Alfalfa'},
    38: {'crop_class': 1, 'h_max': 1.0, 'm_l': 2, 'fr_mid': 1, 'name': 'Camelina'},
    205: {'crop_class': 1, 'h_max': 0.65, 'm_l': 2, 'fr_mid': 1, 'name': 'Triticale'},

    # Crops without custom coefficients
    11: {'crop_class': 1, 'name': 'Tobacco'},
    13: {'crop_class': 1, 'name': 'Pop or Orn Corn'},
    44: {'crop_class': 1, 'name': 'Other Crops'},
    55: {'crop_class': 1, 'name': 'Caneberries'},
    57: {'crop_class': 1, 'name': 'Herbs'},
    59: {'crop_class': 1, 'name': 'Sod/Grass Seed'},
    60: {'crop_class': 1, 'name': 'Switchgrass'},
    219: {'crop_class': 1, 'name': 'Greens'},
    225: {'crop_class': 1, 'name': 'Dbl Crop WinWht/Corn'},
    226: {'crop_class': 1, 'name': 'Dbl Crop Oats/Corn'},
    230: {'crop_class': 1, 'name': 'Dbl Crop Lettuce/Durum Wht'},
    231: {'crop_class': 1, 'name': 'Dbl Crop Lettuce/Cantaloupe'},
    232: {'crop_class': 1, 'name': 'Dbl Crop Lettuce/Cotton'},
    233: {'crop_class': 1, 'name': 'Dbl Crop Lettuce/Barley'},
    234: {'crop_class': 1, 'name': 'Dbl Crop Durum Wht/Sorghum'},
    235: {'crop_class': 1, 'name': 'Dbl Crop Barley/Sorghum'},
    236: {'crop_class': 1, 'name': 'Dbl Crop WinWht/Sorghum'},
    237: {'crop_class': 1, 'name': 'Dbl Crop Barley/Corn'},
    238: {'crop_class': 1, 'name': 'Dbl Crop WinWht/Cotton'},
    239: {'crop_class': 1, 'name': 'Dbl Crop Soybeans/Cotton'},
    240: {'crop_class': 1, 'name': 'Dbl Crop Soybeans/Oats'},
    241: {'crop_class': 1, 'name': 'Dbl Crop Corn/Soybeans'},
    242: {'crop_class': 1, 'name': 'Blueberries'},
    249: {'crop_class': 1, 'name': 'Gourds'},
    250: {'crop_class': 1, 'name': 'Cranberries'},
    254: {'crop_class': 1, 'name': 'Dbl Crop Barley/Soybeans'},

    # Vines
    69: {
        'crop_class': 2, 'h_max': 2, 'm_l': 1.5, 'fr_mid': 0.65, 'fr_end': 0.43,
        'ls_start': 205, 'ls_stop': 265, 'name': 'Wine Grapes'
    },
    78: {
        'crop_class': 2, 'h_max': 2, 'm_l': 1.5, 'fr_mid': 0.95, 'fr_end': 0.51,
        'ls_start': 205, 'ls_stop': 265, 'name': 'Grapes (table/raisin)'
    },

    # Trees
    66: {
        'crop_class': 3, 'h_max': 3, 'm_l': 2, 'fr_mid': 0.95, 'fr_end': 0.75,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Cherries',
    },
    67: {
        'crop_class': 3, 'h_max': 3, 'm_l': 1.5, 'fr_mid': 1.0, 'fr_end': 0.71,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Peaches',
    },
    68: {
        'crop_class': 3, 'h_max': 3, 'm_l': 2, 'fr_mid': 0.95, 'fr_end': 0.75,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Apples',
    },
    71: {
        'crop_class': 3, 'h_max': 4, 'm_l': 1.5, 'fr_mid': 0.89, 'fr_end': 0.63,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Other Tree Crops'
    },
    72: {
        'crop_class': 3, 'h_max': 2.5, 'm_l': 1.5, 'fr_mid': 0.71, 'fr_end': 0.94,
        'ls_start': 270, 'ls_stop': 365, 'name': 'Citrus',
    },
    74: {
        'crop_class': 3, 'h_max': 4, 'm_l': 1.5, 'fr_mid': 0.89, 'fr_end': 0.80,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Pecans',
    },
    75: {
        'crop_class': 3, 'h_max': 4, 'm_l': 1.5, 'fr_mid': 0.81, 'fr_end': 0.59,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Almonds',
    },
    76: {
        'crop_class': 3, 'h_max': 5, 'm_l': 1.5, 'fr_mid': 0.9, 'fr_end': 0.52,
        'ls_start': 250, 'ls_stop': 280, 'name': 'Walnuts',
    },
    77: {
        'crop_class': 3, 'h_max': 3, 'm_l': 2, 'fr_mid': 0.95, 'fr_end': 0.75,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Pears',
    },
    204: {
        'crop_class': 3, 'h_max': 3, 'm_l': 1.5, 'fr_mid': 0.81, 'fr_end': 0.57,
        'ls_start': 200, 'ls_stop': 240, 'name': 'Pistachios',
    },
    210: {
        'crop_class': 3, 'h_max': 3, 'm_l': 1.5, 'fr_mid': 0.95, 'fr_end': 0.71,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Prunes',
    },
    211: {
        'crop_class': 3, 'h_max': 4, 'm_l': 1.5, 'fr_mid': 0.48, 'fr_end': 0.46,
        'ls_start': 240, 'ls_stop': 330, 'name': 'Olives',
    },
    212: {
        'crop_class': 3, 'h_max': 2.5, 'm_l': 1.5, 'fr_mid': 0.71, 'fr_end': 0.94,
        'ls_start': 270, 'ls_stop': 365, 'name': 'Oranges',
    },
    215: {
        'crop_class': 3, 'h_max': 3, 'm_l': 2, 'fr_mid': 0.81, 'fr_end': 0.73,
        'ls_start': 270, 'ls_stop': 365, 'name': 'Avocados',
    },
    218: {
        'crop_class': 3, 'h_max': 3, 'm_l': 1.5, 'fr_mid': 0.95, 'fr_end': 0.75,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Nectarines',
    },
    220: {
        'crop_class': 3, 'h_max': 3, 'm_l': 1.5, 'fr_mid': 0.95, 'fr_end': 0.71,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Plums',
    },
    223: {
        'crop_class': 3, 'h_max': 3, 'm_l': 1.5, 'fr_mid': 0.95, 'fr_end': 0.71,
        'ls_start': 270, 'ls_stop': 300, 'name': 'Apricots',
    },

    # Tree crops without custom coefficients
    70: {'crop_class': 3, 'name': 'Christmas Trees'},
    217: {'crop_class': 3, 'name': 'Pomegranates'},

    # Fallow
    61: {'crop_class': 6, 'name': 'Fallow/Idle Cropland'},

    # Grass pasture
    176: {'crop_class': 7, 'name': 'Grass/pasture'},
}



import base64
import logging
import os

import ee
import pytest


@pytest.fixture(scope="session", autouse=True)
def test_init():
    logging.basicConfig(level=logging.DEBUG, format='%(message)s')
    logging.getLogger('googleapiclient').setLevel(logging.ERROR)
    logging.debug('Test Setup')

    # For GitHub Actions authenticate using private key environment variable
    if 'EE_PRIVATE_KEY_B64' in os.environ:
        print('Writing privatekey.json from environmental variable ...')
        content = base64.b64decode(os.environ['EE_PRIVATE_KEY_B64']).decode('ascii')
        EE_KEY_FILE = 'privatekey.json'
        with open(EE_KEY_FILE, 'w') as f:
            f.write(content)
        ee.Initialize(ee.ServiceAccountCredentials('', key_file=EE_KEY_FILE))
    else:
        ee.Initialize()



import ee
import openet.core.common

from .model import Model
# from . import model
from . import utils
# import utils


def lazy_property(fn):
    """Decorator that makes a property lazy-evaluated

    https://stevenloria.com/lazy-properties/
    """
    attr_name = '_lazy_' + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazy_property


class Image():
    """GEE based model for computing SIMS ETcb"""

    def __init__(
        self,
        image,
        crop_type_source='USDA/NASS/CDL',
        crop_type_remap='CDL',
        crop_type_kc_flag=False,
        crop_type_annual_skip_flag=False,
        et_reference_source=None,
        et_reference_band=None,
        et_reference_factor=None,
        et_reference_resample=None,
        mask_non_ag_flag=False,
        water_kc_flag=True,
        reflectance_type='SR',
    ):
        """Earth Engine based SIMS image object

        Parameters
        ----------
        image : ee.Image
            Required band: ndvi
            Required properties: system:time_start, system:index, system:id
        crop_type_source : str, optional
            Crop type source.  The default is the Cropland Data Layer (CDL) assets.
            Source should be an EE image or collection ID (or ee.Image).
            Currently only the OpenET crop type and CDL images are supported.
        crop_type_remap : {'CDL'}, optional
            Currently only CDL crop type values are supported.
        crop_type_kc_flag : bool, optional
            If True, compute Kc using crop type specific coefficients.
            If False, use generic crop class coefficients.
            The default is False.
        crop_type_annual_skip_flag : bool, optional
            If True, the crop type specific coefficients are NOT used for annual crops.
            If False, the crop type specific coefficients are used for annual crops.
            This flag is only applied/used if crop_type_kc_flag is also True.
            The default is False.
        et_reference_source : str, float, optional
            Reference ET source (the default is None).
            Parameter is required if computing 'et' or 'et_reference'.
        et_reference_band : str, optional
            Reference ET band name (the default is None).
            Parameter is required if computing 'et' or 'et_reference'.
        et_reference_factor : float, None, optional
            Reference ET scaling factor.  The default is None which is
            equivalent to 1.0 (or no scaling).
        et_reference_resample : {'nearest', 'bilinear', 'bicubic', None}, optional
            Reference ET resampling.  The default is None which is equivalent
            to nearest neighbor resampling.
        mask_non_ag_flag : bool, optional
            If True, mask all pixels that don't map to a crop_class.
            The default is False.
        water_kc_flag : bool, optional
            If True, set Kc for water pixels to 1.05.  The default is True.
        reflectance_type : {'SR', 'TOA'}, optional
            Used to select the fractional cover equation (the default is 'SR').

        Notes
        -----
        Fc = (NDVI * 1.26) - 0.18
        Kc = f(Fc) [based on crop type or crop class]
        ETcb = Kc * ETo

        References
        ----------
        .. [1] Johnson, L. and T. Trout, 2012. Satellite NDVI assisted
            monitoring of vegetable crop evapotranspiration in California's San
            Joaquin Valley. Remote Sensing 4:439-455. [EQN 1]

        """
        self.image = image

        # Get system properties from the input image
        self._id = self.image.get('system:id')
        self._index = self.image.get('system:index')
        self._time_start = self.image.get('system:time_start')
        self._properties = {
            'system:index': self._index,
            'system:time_start': self._time_start,
            'image_id': self._id,
        }

        # Build date properties from the system:time_start
        self._date = ee.Date(self._time_start)
        self._year = ee.Number(self._date.get('year'))
        self._start_date = ee.Date(utils.date_0utc(self._date).millis())
        self._end_date = self._start_date.advance(1, 'day')
        self._doy = self._date.getRelative('day', 'year').add(1).int()

        # Reference ET parameters
        self.et_reference_source = et_reference_source
        self.et_reference_band = et_reference_band
        self.et_reference_factor = et_reference_factor
        self.et_reference_resample = et_reference_resample

        # Check reference ET parameters
        if et_reference_factor and not utils.is_number(et_reference_factor):
            raise ValueError('et_reference_factor must be a number')
        if et_reference_factor and self.et_reference_factor < 0:
            raise ValueError('et_reference_factor must be greater than zero')
        et_reference_resample_methods = ['nearest', 'bilinear', 'bicubic']
        if (et_reference_resample and
                et_reference_resample.lower() not in et_reference_resample_methods):
            raise ValueError('unsupported et_reference_resample method')

        self.reflectance_type = reflectance_type

        # CGM - Model class could inherit these from Image instead of passing them
        #   Could pass time_start instead of separate year and doy
        self.model = Model(
            year=self._year,
            doy=self._doy,
            crop_type_source=crop_type_source,
            crop_type_remap=crop_type_remap,
            crop_type_kc_flag=crop_type_kc_flag,
            crop_type_annual_skip_flag=crop_type_annual_skip_flag,
            mask_non_ag_flag=mask_non_ag_flag,
            water_kc_flag=water_kc_flag,
            reflectance_type=reflectance_type,
        )

    def calculate(self, variables=['et']):
        """Return a multiband image of calculated variables

        Parameters
        ----------
        variables : list

        Returns
        -------
        ee.Image

        """
        output_images = []
        for v in variables:
            if v.lower() == 'et':
                output_images.append(self.et.float())
            elif v.lower() == 'et_reference':
                output_images.append(self.et_reference.float())
            elif v.lower() == 'et_fraction':
                output_images.append(self.et_fraction.float())
            # elif v.lower() == 'crop_class':
            #     output_images.append(self.crop_class)
            # elif v.lower() == 'crop_type':
            #     output_images.append(self.crop_type)
            elif v.lower() == 'fc':
                output_images.append(self.fc.float())
            elif v.lower() == 'kc':
                output_images.append(self.kc.float())
            elif v.lower() == 'mask':
                output_images.append(self.mask)
            elif v.lower() == 'ndvi':
                output_images.append(self.ndvi.float())
            elif v.lower() == 'time':
                output_images.append(self.time)
            else:
                raise ValueError(f'unsupported variable: {v}')

        return ee.Image(output_images).set(self._properties)

    @lazy_property
    def et_fraction(self):
        """Fraction of reference ET (equivalent to the Kc)

        This method is basically identical to the "kc" method and is only
        provided to simplify interaction with the interpolation tools in the
        Collection.

        Returns
        -------
        ee.Image

        """
        return self.kc.rename(['et_fraction']).set(self._properties)
        # ET fraction could also be calculated from ET and ET reference
        # return self.et.divide(self.et_reference)\
        #     .rename(['et_fraction']).set(self._properties)

    @lazy_property
    def et_reference(self):
        """Reference ET for the image date

        Returns
        -------
        ee.Image

        """
        if utils.is_number(self.et_reference_source):
            # Interpret numbers as constant images
            # CGM - Should we use the ee_types here instead?
            #   i.e. ee.ee_types.isNumber(self.et_reference_source)
            et_reference_img = ee.Image.constant(self.et_reference_source)
        elif type(self.et_reference_source) is str:
            # Assume a string source is an image collection ID (not an image ID)
            et_reference_coll = (
                ee.ImageCollection(self.et_reference_source)
                .filterDate(self._start_date, self._end_date)
                .select([self.et_reference_band])
            )
            et_reference_img = ee.Image(et_reference_coll.first())
            if self.et_reference_resample in ['bilinear', 'bicubic']:
                et_reference_img = et_reference_img.resample(self.et_reference_resample)
        else:
            raise ValueError(
                f'unsupported et_reference_source: {self.et_reference_source}'
            )

        if self.et_reference_factor:
            et_reference_img = et_reference_img.multiply(self.et_reference_factor)

        return (
            self.ndvi.multiply(0).add(et_reference_img)
            .rename(['et_reference']).set(self._properties)
        )

    @lazy_property
    def et(self):
        """Actual ET (ETcb)

        Returns
        -------
        ee.Image

        """
        return self.kc.multiply(self.et_reference).rename(['et']).set(self._properties)

    @lazy_property
    def crop_class(self):
        """Generic crop classes

        Returns
        -------
        ee.Image

        """
        # Map the the crop class values to the NDVI image
        return (
            self.ndvi.multiply(0).add(self.model.crop_class)
            .rename('crop_class').set(self._properties)
        )

    @lazy_property
    def crop_type(self):
        """Crop type

        Returns
        -------
        ee.Image

        """
        # Map the the crop class values to the NDVI image
        # Crop type image ID property is set in model function
        return self.ndvi.multiply(0).add(self.model.crop_type).rename(['crop_type'])

    @lazy_property
    def fc(self):
        """Fraction of cover (fc)

        Returns
        -------
        ee.Image

        """
        return self.model.fc(self.ndvi).rename(['fc']).set(self._properties)

    @lazy_property
    def kc(self):
        """Crop coefficient (Kc)

        Returns
        -------
        ee.Image

        """
        return self.model.kc(self.ndvi).rename(['kc']).set(self._properties)

    @lazy_property
    def mask(self):
        """Mask of all active pixels based on the final Kc

        Using Kc here to capture any masking that might be in the crop_type

        Returns
        -------
        ee.Image

        """
        return (
            self.kc.multiply(0).add(1).updateMask(1)
            .rename(['mask']).set(self._properties).uint8()
        )

    @lazy_property
    def ndvi(self):
        """Normalized difference vegetation index (NDVI)

        Returns
        -------
        ee.Image

        """
        return self.image.select(['ndvi']).set(self._properties)

    # @lazy_property
    # def quality(self):
    #     """Set quality to 1 for all active pixels (for now)"""
    #     return self.mask\
    #         .rename(['quality']).set(self._properties)

    @lazy_property
    def time(self):
        """Image of the 0 UTC time (in milliseconds)

        Returns
        -------
        ee.Image

        """
        return (
            self.mask
            .double().multiply(0).add(utils.date_0utc(self._date).millis())
            .rename(['time']).set(self._properties)
        )

    @classmethod
    def from_image_id(cls, image_id, **kwargs):
        """Construct a SIMS Image instance from an image ID

        Parameters
        ----------
        image_id : str
            A full earth engine image ID.
            (i.e. 'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716')
        kwargs : dict
            Keyword arguments to pass through to model init.

        Returns
        -------
        new instance of Image class

        Raises
        ------
        ValueError for an unsupported collection ID.

        """
        collection_methods = {
            'LANDSAT/LC09/C02/T1_L2': 'from_landsat_c2_sr',
            'LANDSAT/LC08/C02/T1_L2': 'from_landsat_c2_sr',
            'LANDSAT/LE07/C02/T1_L2': 'from_landsat_c2_sr',
            'LANDSAT/LT05/C02/T1_L2': 'from_landsat_c2_sr',
            'LANDSAT/LT04/C02/T1_L2': 'from_landsat_c2_sr',
        }

        try:
            method_name = collection_methods[image_id.rsplit('/', 1)[0]]
        except KeyError:
            raise ValueError(f'unsupported collection ID: {image_id}')
        except Exception as e:
            raise Exception(f'unhandled exception: {e}')

        method = getattr(Image, method_name)

        return method(ee.Image(image_id), **kwargs)

    @classmethod
    def from_landsat_c2_sr(cls, sr_image, cloudmask_args={}, **kwargs):
        """Construct a SIMS Image instance from a Landsat C02 level 2 (SR) image

        Parameters
        ----------
        sr_image : ee.Image, str
            A raw Landsat Collection 2 level 2 (SR) image or image ID.
        cloudmask_args : dict
            keyword arguments to pass through to cloud mask function
        kwargs : dict
            Keyword arguments to pass through to model init.

        Returns
        -------
        new instance of Image class

        Notes
        -----
        https://www.usgs.gov/faqs/how-do-i-use-a-scale-factor-landsat-level-2-science-products?qt-news_science_products=0#qt-news_science_products
        https://www.usgs.gov/core-science-systems/nli/landsat/landsat-collection-2-level-2-science-products

        """
        sr_image = ee.Image(sr_image)

        # Use the SPACECRAFT_ID property identify each Landsat type
        spacecraft_id = ee.String(sr_image.get('SPACECRAFT_ID'))

        # Rename bands to generic names
        input_bands = ee.Dictionary({
            'LANDSAT_4': ['SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B7',
                          'ST_B6', 'QA_PIXEL', 'QA_RADSAT'],
            'LANDSAT_5': ['SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B7',
                          'ST_B6', 'QA_PIXEL', 'QA_RADSAT'],
            'LANDSAT_7': ['SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B7',
                          'ST_B6', 'QA_PIXEL', 'QA_RADSAT'],
            'LANDSAT_8': ['SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7',
                          'ST_B10', 'QA_PIXEL', 'QA_RADSAT'],
            'LANDSAT_9': ['SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7',
                          'ST_B10', 'QA_PIXEL', 'QA_RADSAT'],
        })
        output_bands = ['blue', 'green', 'red', 'nir', 'swir1', 'swir2',
                        'lst', 'QA_PIXEL', 'QA_RADSAT']
        prep_image = (
            sr_image
            .select(input_bands.get(spacecraft_id), output_bands)
            .multiply([0.0000275, 0.0000275, 0.0000275, 0.0000275, 0.0000275, 0.0000275, 0.00341802, 1, 1])
            .add([-0.2, -0.2, -0.2, -0.2, -0.2, -0.2, 149.0, 0, 0])
        )

        # Default the cloudmask flags to True if they were not
        # Eventually these will probably all default to True in openet.core
        if 'cirrus_flag' not in cloudmask_args.keys():
            cloudmask_args['cirrus_flag'] = True
        if 'dilate_flag' not in cloudmask_args.keys():
            cloudmask_args['dilate_flag'] = True
        if 'shadow_flag' not in cloudmask_args.keys():
            cloudmask_args['shadow_flag'] = True
        if 'snow_flag' not in cloudmask_args.keys():
            cloudmask_args['snow_flag'] = True
        if 'cloud_score_flag' not in cloudmask_args.keys():
            cloudmask_args['cloud_score_flag'] = False
        if 'cloud_score_pct' not in cloudmask_args.keys():
            cloudmask_args['cloud_score_pct'] = 100
        if 'filter_flag' not in cloudmask_args.keys():
            cloudmask_args['filter_flag'] = False
        if 'saturated_flag' not in cloudmask_args.keys():
            cloudmask_args['saturated_flag'] = False

        cloud_mask = openet.core.common.landsat_c2_sr_cloud_mask(sr_image, **cloudmask_args)

        # Build the input image
        # Eventually send the QA band or a cloud mask through also
        input_image = ee.Image([cls._ndvi(prep_image)])

        # Apply the cloud mask and add properties
        input_image = (
            input_image
            .updateMask(cloud_mask)
            .set({
                'system:index': sr_image.get('system:index'),
                'system:time_start': sr_image.get('system:time_start'),
                'system:id': sr_image.get('system:id'),
            })
        )

        return cls(input_image, reflectance_type='SR', **kwargs)

    @staticmethod
    def _ndvi(landsat_image):
        """Normalized difference vegetation index

        Parameters
        ----------
        landsat_image : ee.Image
            "Prepped" Landsat image with standardized band names.

        Returns
        -------
        ee.Image

        """
        return landsat_image.normalizedDifference(['nir', 'red']).rename(['ndvi'])



import datetime
# import pprint

import ee
import pytest

import openet.sims.utils as utils


def test_ee_init():
    assert utils.getinfo(ee.Number(1)) == 1


def test_constant_image_value(expected=10.123456789, tol=0.000001):
    output = utils.constant_image_value(ee.Image.constant(expected))
    assert abs(output['constant'] - expected) <= tol


def test_constant_image_value_band_name(expected=10.123456789, tol=0.000001):
    """Test that a custom band name is carried through"""
    input_img = ee.Image.constant(expected).rename('foo')
    output = utils.constant_image_value(input_img)
    assert abs(output['foo'] - expected) <= tol


def test_constant_image_value_multiband(expected=10.123456789, tol=0.000001):
    """Test that a multiband image returns multiple values"""
    input_img = ee.Image.constant([expected, expected + 1])
    output = utils.constant_image_value(input_img)
    assert abs(output['constant_0'] - expected) <= tol
    assert abs(output['constant_1'] - (expected + 1)) <= tol


def test_constant_image_value_multiband_bands(expected=10.123456789, tol=0.000001):
    """Test that the band names are carried through on a multiband image"""
    input_img = ee.Image.constant([expected, expected + 1]).rename(['foo', 'bar'])
    output = utils.constant_image_value(input_img)
    assert abs(output['foo'] - expected) <= tol
    assert abs(output['bar'] - (expected + 1)) <= tol


@pytest.mark.parametrize(
    'image_id, xy, scale, expected, tol',
    [
        ['USGS/3DEP/10m', [-106.03249, 37.17777], 30, 2364.169, 0.001],
        ['USGS/3DEP/10m', [-106.03249, 37.17777], 10, 2364.138, 0.001],
        ['USGS/3DEP/10m', [-106.03249, 37.17777], 1, 2364.138, 0.001],
        ['NASA/NASADEM_HGT/001', [-106.03249, 37.17777], 30, 2361, 0.001],
    ]
)
def test_point_image_value(image_id, xy, scale, expected, tol):
    output = utils.point_image_value(
        ee.Image(image_id).select(['elevation'], ['output']), xy, scale
    )
    assert abs(output['output'] - expected) <= tol


@pytest.mark.parametrize(
    'image_id, image_date, xy, scale, expected, tol',
    [
        ['USGS/3DEP/10m', '2012-04-04', [-106.03249, 37.17777], 30, 2364.169, 0.001],
        ['USGS/3DEP/10m', '2012-04-04', [-106.03249, 37.17777], 10, 2364.097, 0.001],
        ['USGS/3DEP/10m', '2012-04-04', [-106.03249, 37.17777], 1, 2364.138, 0.001],
        ['NASA/NASADEM_HGT/001', '2012-04-04', [-106.03249, 37.17777], 30, 2361, 0.001],
    ]
)
def test_point_coll_value(image_id, image_date, xy, scale, expected, tol):
    # The image must have a system:time_start for this function to work correctly
    input_img = (
        ee.Image(image_id).select(['elevation'], ['output'])
        .set({'system:time_start': ee.Date(image_date).millis()})
    )
    output = utils.point_coll_value(ee.ImageCollection([input_img]), xy, scale)
    assert abs(output['output'][image_date] - expected) <= tol


@pytest.mark.parametrize(
    'input, expected',
    [
        ['2015-07-13T18:33:39', 1436745600000],
        ['2015-07-13T00:00:00', 1436745600000],

    ]
)
def test_date_0utc(input, expected):
    assert utils.getinfo(utils.date_0utc(ee.Date(input)).millis()) == expected


@pytest.mark.parametrize(
    # Note: These are made up values
    'input, expected',
    [
        [300, True],
        ['300', True],
        [300.25, True],
        ['300.25', True],
        ['a', False],
    ]
)
def test_is_number(input, expected):
    assert utils.is_number(input) == expected


def test_millis():
    assert utils.millis(datetime.datetime(2015, 7, 13)) == 1436745600000


def test_valid_date():
    assert utils.valid_date('2015-07-13') is True
    assert utils.valid_date('2015-02-30') is False
    assert utils.valid_date('20150713') is False
    assert utils.valid_date('07/13/2015') is False
    assert utils.valid_date('07-13-2015', '%m-%d-%Y') is True



import calendar
import datetime
import logging
from time import sleep

import ee


def getinfo(ee_obj, n=4):
    """Make an exponential back off getInfo call on an Earth Engine object"""
    output = None
    for i in range(1, n):
        try:
            output = ee_obj.getInfo()
        except ee.ee_exception.EEException as e:
            logging.info(f'    Resending query ({i}/{n})')
            logging.info(f'    {e}')
            sleep(i ** 3)
            # if ('Earth Engine memory capacity exceeded' in str(e) or
            #         'Earth Engine capacity exceeded' in str(e)):
            #     logging.info(f'    Resending query ({i}/{n})')
            #     logging.debug(f'    {e}')
            #     sleep(i ** 2)
            # else:
            #     raise e
        except Exception as e:
            logging.info(e)
            break

        if output:
            break

    return output


# TODO: Import from openet.core.utils instead of defining here
def constant_image_value(image, crs='EPSG:32613', scale=1):
    """Extract the output value from a calculation done with constant images"""
    rr_params = {
        'reducer': ee.Reducer.first(),
        'geometry': ee.Geometry.Rectangle([0, 0, 10, 10], crs, False),
        'scale': scale,
    }
    return getinfo(ee.Image(image).reduceRegion(**rr_params))


def point_image_value(image, xy, scale=1):
    """Extract the output value from a calculation at a point"""
    rr_params = {
        'reducer': ee.Reducer.first(),
        'geometry': ee.Geometry.Point(xy),
        'scale': scale,
    }
    return getinfo(ee.Image(image).reduceRegion(**rr_params))


def point_coll_value(coll, xy, scale=1):
    """Extract the output value from a calculation at a point"""
    output = getinfo(coll.getRegion(ee.Geometry.Point(xy), scale=scale))

    # Structure output to easily be converted to a Pandas dataframe
    # First key is band name, second key is the date string
    col_dict = {}
    info_dict = {}
    for i, k in enumerate(output[0][4:]):
        col_dict[k] = i + 4
        info_dict[k] = {}
    for row in output[1:]:
        date = datetime.datetime.utcfromtimestamp(row[3] / 1000.0).strftime('%Y-%m-%d')
        for k, v in col_dict.items():
            info_dict[k][date] = row[col_dict[k]]

    return info_dict
    # return pd.DataFrame.from_dict(info_dict)


def date_0utc(date):
    """Get the 0 UTC date for a date

    Parameters
    ----------
    date : ee.Date

    Returns
    -------
    ee.Date

    """
    return ee.Date.fromYMD(date.get('year'), date.get('month'), date.get('day'))


def is_number(x):
    try:
        float(x)
        return True
    except:
        return False


def millis(input_dt):
    """Convert datetime to milliseconds since epoch

    Parameters
    ----------
    input_dt : datetime

    Returns
    -------
    int

    """
    return 1000 * int(calendar.timegm(input_dt.timetuple()))


def valid_date(date_str, date_fmt='%Y-%m-%d'):
    """Check if a datetime can be built from a date string and if it is valid"""
    try:
        datetime.datetime.strptime(date_str, date_fmt)
        return True
    except:
        return False



import ee
import pytest

import openet.sims.data as data
import openet.sims.utils as utils


def test_int_scalar():
    assert data.int_scalar == 100
    # assert data.int_scalar
    # assert data.int_scalar % 10 == 0


def test_cdl_dict():
    assert type(data.cdl) is dict


# CGM - This doesn't really need to be 4 separate tests
#   It could probably be rewritten to make a single call on the collection
# @pytest.mark.parametrize('year', [2016, 2017, 2018, 2019])
@pytest.mark.parametrize('year', [2019])
def test_cdl_crop_types(year):
    output = list(map(round, utils.getinfo(
        ee.Image(f'USDA/NASS/CDL/{year}').get('cropland_class_values'))))
    for crop_type, crop_data in data.cdl.items():
        # Crop type 78 is non-standard CDL code being used for Grapes (table/raisin)
        if crop_type == 78:
            continue
        assert crop_type in output
    # assert all(crop_type in output for crop_type, crop_data in data.cdl.items())

# CGM - Testing to see if crop type keys have changed or all present doesn't
#   seem like a very useful test.
# def test_cdl_crop_types():
# #     # CDL codes in data file:
# #     assert set(data.cdl.keys()) == set([
# #         1, 2, 3, 4, 5, 6, 12, 14, 21, 23, 24, 27, 28, 29, 31, 32, 33, 36,
# #         41, 42, 43, 46, 48, 49, 51, 52, 53, 54, 58,
# #         66, 67, 68, 69, 72, 75, 76, 77, 141, 142,
# #         204, 206, 207, 208, 209, 211, 212, 214, 221, 222, 223, 227, 229,
# #         243, 244, 245, 246, 247, 248])
# #     # CDL codes not in data file:
# #     #   10, 11, 13, 22, 25, 26, 30, 32, 34, 35, 37, 38, 39, 44, 45, 47,
# #     #   50, 55, 56, 57, 59, 61, 70, 71, 74,
# #     #   205, 210, 213, 215, 216, 217, 218, 219, 220, 224, 225, 226,
# #     #   230, 231, 232, 233, 234, 235, 236, 237, 238, 239,
# #     #   240, 241, 242, 249, 250, 254


@pytest.mark.parametrize('param', ['crop_class', 'h_max', 'm_l', 'fr_mid'])
def test_cdl_parameters(param):
    # Check that all default parameter keys have a value
    # Crops without a key will use general Kc equation
    # CGM - This test isn't very informative about which crop is the problem
    assert all(crop_data[param] for crop_data in data.cdl.values()
               if param in crop_data.keys())


@pytest.mark.parametrize('param', ['fr_end', 'ls_start', 'ls_stop'])
def test_cdl_class3_parameters(param):
    assert all(crop_data[param] for crop_data in data.cdl.values()
               if crop_data['crop_class'] == 3 and param in crop_data.keys())


@pytest.mark.parametrize(
    'crop_type, crop_class',
    [
        [1, 1],
        [69, 2],
        [66, 3],
        [3, 5],    # Rice was switched to class 5 instead of 1
        [61, 6],   # Fallow was switched to class 6 instead of 1
        [176, 7],  # Grass/pasture was switched to class 7 instead of 1
    ]
)
def test_cdl_crop_classes(crop_type, crop_class):
    assert data.cdl[crop_type]['crop_class'] == crop_class


# CGM - How would I do this?
# def test_cdl_int_scalar_digits():
#     # Check if any values have more digits than can be handled by int_scalar
#     assert False
